from __future__ import print_function, division
import numpy as np
import fipy
from fipy.terms import TransientTerm, DiffusionTerm, ImplicitSourceTerm
from ciabatta.meta import make_repr_str


class Field(object):
    def __init__(self, mesh, c_0):
        self.mesh = mesh
        self.c_0 = c_0
        self.c = fipy.CellVariable(mesh=self.mesh, value=self.c_0)

    @property
    def dim(self):
        return self.mesh.dim

    def _ps_to_rs(self, ps):
        return ps.r_w.T

    def _get_nearest_cell_ids(self, rs):
        return self.mesh._getNearestCellID(rs)

    def get_nearest_cell_ids(self, ps):
        return self._get_nearest_cell_ids(self._ps_to_rs(ps))

    def _get_grad_i(self, rs):
        return self.c.grad[:, self._get_nearest_cell_ids(rs)].value.T

    def _get_val_i(self, rs):
        return self.c[:, self._get_nearest_cell_ids(rs)].value.T

    def get_grad_i(self, ps):
        return self._get_grad_i(self._ps_to_rs(ps))

    def get_val_i(self, ps):
        return self._get_val_i(self._ps_to_rs(ps))

    def __repr__(self):
        fs = [('dim', self.dim), ('mesh', self.mesh), ('c_0', self.c_0)]
        return make_repr_str(self, fs)


class FoodField(Field):

    def __init__(self, mesh, D, delta, c_0):
        super(FoodField, self).__init__(mesh, c_0)
        self.D = D
        self.delta = delta

        self.rho = fipy.CellVariable(mesh=self.mesh, value=0.0)

        self.eq = (TransientTerm(var=self.c) ==
                   DiffusionTerm(coeff=self.D, var=self.c) -
                   ImplicitSourceTerm(coeff=self.delta * self.rho, var=self.c))

    def _get_rho_array(self, ps):
        rho_array = np.zeros(self.rho.shape)
        cids = self.get_nearest_cell_ids(ps)
        for i in cids:
            rho_array[i] += 1.0 / self.rho.mesh.cellVolumes[i]
        return rho_array

    def iterate(self, ps, dt):
        rho_array = self._get_rho_array(ps)
        self.rho.setValue(rho_array)
        self.eq.solve(dt=dt)

    def __repr__(self):
        fs = [('dim', self.dim), ('mesh', self.mesh), ('c_0', self.c_0),
              ('D', self.D), ('delta', self.delta)]
        return make_repr_str(self, fs)


class NoneFoodField(object):

    def iterate(self, ps, dt):
        return

    def __repr__(self):
        fs = []
        return make_repr_str(self, fs)


def food_field_factory(c_field_flag, L, c_dx, c_D, c_delta, c_0, obstructor):
    if c_field_flag:
        mesh = obstructor.get_mesh(L, c_dx)
        c_field = FoodField(mesh, c_D, c_delta, c_0)
    else:
        c_field = NoneFoodField()
    return c_field



from __future__ import print_function, division
from ahoy.rudders import TumbleRudders, RotationRudders, rudders_factory
from ciabatta.meta import make_repr_str


class RudderSets(object):

    def __init__(self):
        self.sets = []

    @property
    def chi(self):
        for rudders in self.sets:
            if rudders.is_chemotactic:
                return rudders.chi

    @property
    def does_tumbling(self):
        for rs in self.sets:
            if isinstance(rs, TumbleRudders):
                return True
        return False

    @property
    def does_rotation(self):
        for rs in self.sets:
            if isinstance(rs, RotationRudders):
                return True
        return False

    @property
    def chemo_rudders(self):
        for rs in self.sets:
            if rs.is_chemotactic:
                return rs
        raise Exception

    def rotate(self, directions, dt, rng):
        for rudders in self.sets:
            directions = rudders.rotate(directions, dt, rng)
        return directions

    def __repr__(self):
        fs = [('sets', self.sets)]
        return make_repr_str(self, fs)


def rudder_set_factory(temporal_chemo_flag,
                       ds,
                       ps, v_0, dt_mem, t_mem, time,
                       c_field_flag, c_field,
                       onesided_flag, chi,
                       tumble_flag, p_0, tumble_chemo_flag,
                       rotation_flag, Dr_0, dim, rotation_chemo_flag):
    D_rot_0 = 0.0
    if tumble_flag:
        D_rot_0 += p_0
    if rotation_flag:
        D_rot_0 += Dr_0
    t_rot_0 = 1.0 / D_rot_0

    sets = RudderSets()
    if tumble_flag:
        rudders = rudders_factory(True, dim,
                                  tumble_chemo_flag,
                                  p_0,
                                  onesided_flag, chi,
                                  temporal_chemo_flag,
                                  ds,
                                  ps, v_0, dt_mem, t_mem, t_rot_0, time,
                                  c_field_flag, c_field)
        sets.sets.append(rudders)
    if rotation_flag:
        rudders = rudders_factory(False, dim,
                                  rotation_chemo_flag,
                                  Dr_0,
                                  onesided_flag, chi,
                                  temporal_chemo_flag,
                                  ds,
                                  ps, v_0, dt_mem, t_mem, t_rot_0, time,
                                  c_field_flag, c_field)
        sets.sets.append(rudders)
    return sets



from __future__ import print_function, division
import sys
import matplotlib.pyplot as plt
from ahoy.plot import plot


def plot_chi_uds_x():
    ax = plt.gca()
    plot.plot_chi_uds_x(sys.argv[1:], ax)
    plt.show()


def plot_Dr_0_Ds_scalar():
    ax = plt.gca()
    plot.plot_Dr_0_Ds_scalar(sys.argv[1:], ax)
    plt.show()


def plot_linear_density():
    ax = plt.gca()
    plot.plot_linear_density(sys.argv[1], ax)
    plt.show()


def plot_p_0_Ds_scalar():
    ax = plt.gca()
    plot.plot_p_0_Ds_scalar(sys.argv[1:], ax)
    plt.show()


def plot_pf_Ds_scalar():
    ax = plt.gca()
    plot.plot_pf_Ds_scalar(sys.argv[1:], ax)
    plt.show()


def plot_pf_uds_x():
    ax = plt.gca()
    plot.plot_pf_uds_x(sys.argv[1:], ax)
    plt.show()


def plot_t_Ds_scalar():
    ax = plt.gca()
    plot.plot_t_Ds_scalar(sys.argv[1], ax)
    plt.show()


def plot_t_Ds_vector():
    ax = plt.gca()
    plot.plot_t_Ds_vector(sys.argv[1], ax)
    plt.show()


def plot_t_rs_scalar():
    ax = plt.gca()
    plot.plot_t_rs_scalar(sys.argv[1], ax)
    plt.show()


def plot_t_rs_vector():
    ax = plt.gca()
    plot.plot_t_rs_vector(sys.argv[1], ax)
    plt.show()


def plot_t_u_nets_scalar():
    ax = plt.gca()
    plot.plot_t_u_nets_scalar(sys.argv[1], ax)
    plt.show()


def plot_t_u_nets_vector():
    ax = plt.gca()
    plot.plot_t_u_nets_vector(sys.argv[1], ax)
    plt.show()


def plot_t_uds_scalar():
    ax = plt.gca()
    plot.plot_t_uds_scalar(sys.argv[1], ax)
    plt.show()


def plot_t_uds_vector():
    ax = plt.gca()
    plot.plot_t_uds_vector(sys.argv[1], ax)
    plt.show()


def plot_vis():
    plot.plot_vis(sys.argv[1])



from __future__ import print_function, division
import numpy as np
from ahoy import stime, directions, c_measurers, dc_dx_measurers
import test


class TestLinearSpatialDcDxMeasurer(test.TestBase):

    def setUp(self):
        super(TestLinearSpatialDcDxMeasurer, self).setUp()

    def run_nd(self, dim, u_0, dc_dxs_expected):
        ds = directions.directions_nd(u_0)
        grad_c_measurer = c_measurers.ConstantGradCMeasurer(self.n, dim)
        dc_dx_measurer = dc_dx_measurers.SpatialDcDxMeasurer(ds,
                                                             grad_c_measurer)
        dc_dxs = dc_dx_measurer.get_dc_dxs()
        self.assertTrue(np.allclose(dc_dxs, dc_dxs_expected))

    def run_parallel_nd(self, dim):
        u_0 = np.zeros([self.n, dim])
        u_0[:, 0] = 1.0
        self.run_nd(dim, u_0, 1.0)

    def run_antiparallel_nd(self, dim):
        u_0 = np.zeros([self.n, dim])
        u_0[:, 0] = -1.0
        self.run_nd(dim, u_0, -1.0)

    def run_perp_nd(self, dim):
        u_0 = np.zeros([self.n, dim])
        u_0[:, 1] = -1.0
        self.run_nd(dim, u_0, 0.0)

    def test_parallel_2d(self):
        self.run_parallel_nd(2)

    def test_parallel_1d(self):
        self.run_parallel_nd(1)

    def test_antiparallel_2d(self):
        self.run_antiparallel_nd(2)

    def test_antiparallel_1d(self):
        self.run_antiparallel_nd(1)

    def test_perp_2d(self):
        self.run_perp_nd(2)


class MockPositions(object):
    def __init__(self, dim, n, v_0, u_0):
        self.r = np.zeros([n, dim])
        self.v_0 = v_0
        self.u_0 = u_0

    def iterate(self, dt):
        self.r += self.v_0 * self.u_0 * dt

    @property
    def dr(self):
        return self.r


class TestLinearTemporalDcDxMeasurer(TestLinearSpatialDcDxMeasurer):

    def setUp(self):
        super(TestLinearTemporalDcDxMeasurer, self).setUp()
        self.v_0 = 2.2
        self.dt = 0.01
        self.dt_mem = 0.05
        self.t_mem = 5.0
        self.t_rot_0 = 1.0

    def run_nd(self, dim, u_0, dc_dxs_expected):
        time = stime.Time()
        ps = MockPositions(dim, self.n, self.v_0, u_0)
        c_measurer = c_measurers.LinearCMeasurer(ps)
        dc_dx_measurer = dc_dx_measurers.TemporalDcDxMeasurer(c_measurer,
                                                              self.v_0,
                                                              self.dt_mem,
                                                              self.t_mem,
                                                              self.t_rot_0,
                                                              time)
        while time.t < 2.0 * self.t_mem:
            ps.iterate(self.dt)
            dc_dxs = dc_dx_measurer.get_dc_dxs()
            time.iterate(self.dt)
        self.assertTrue(np.allclose(dc_dxs, dc_dxs_expected))



from __future__ import print_function, division
import numpy as np
from ahoy.utils import utils
import test


class TestAngleDensity(test.TestBase):

    def test_angle_density(self):
        ths = np.zeros([1000])
        dth = 0.1
        densities, th_bins = utils.angle_density(ths, dth)
        dth = th_bins[1] - th_bins[0]
        self.assertTrue(np.allclose((densities * dth).sum(), 1.0))



from fipy.tools import numerix
from matplotlib.collections import PolyCollection


class VarPlot(object):

    def __init__(self, var, cmap, axes):
        self.axes = axes
        self.var = var
        self.cmap = cmap
        self.mesh = self.var.mesh
        self.plot_mesh()

    def plot_mesh(self):
        vertexIDs = self.mesh._orderedCellVertexIDs
        vertexCoords = self.mesh.vertexCoords
        xCoords = numerix.take(vertexCoords[0], vertexIDs)
        yCoords = numerix.take(vertexCoords[1], vertexIDs)
        polys = []
        for x, y in zip(xCoords.swapaxes(0, 1), yCoords.swapaxes(0, 1)):
            polys.append(zip(x, y))
        self.collection = PolyCollection(polys)
        self.collection.set_linewidth(0.5)
        self.axes.add_collection(self.collection)
        self.update(self.var)

    def update(self, var):
        rgba = self.cmap(var.value)
        self.collection.set_facecolors(rgba)
        self.collection.set_edgecolors(rgba)



from __future__ import print_function, division
import numpy as np
from ciabatta.meta import make_repr_str
from spatious import vector
from fealty import fields


class Positions(object):

    def __init__(self, r_0):
        self.r = r_0
        self.r_0 = self.r.copy()

    @property
    def n(self):
        return self.r.shape[0]

    @property
    def dim(self):
        return self.r.shape[1]

    @property
    def r_mag(self):
        return vector.vector_mag(self.r)

    @property
    def dr(self):
        return self.r - self.r_0

    @property
    def dr_mag(self):
        return vector.vector_mag(self.dr)

    @property
    def r_w(self):
        return self.r

    @property
    def r_w_mag(self):
        return vector.vector_mag(self.r_w)

    def __repr__(self):
        fs = [('n', self.n), ('dim', self.dim)]
        return make_repr_str(self, fs)


class PeriodicPositions(Positions):

    def __init__(self, L, r_0):
        super(PeriodicPositions, self).__init__(r_0)
        self.L = L

    @property
    def volume(self):
        return np.product(self.L)

    @property
    def r_w(self):
        wraps = self.get_wraps()
        r_w = self.r.copy()
        for i_dim in np.where(np.isfinite(self.L))[0]:
            r_w[:, i_dim] -= wraps[:, i_dim] * self.L[i_dim]
        return r_w

    def get_density_field(self, dx):
        return fields.density(self.r_w, self.L, dx)

    def get_wraps(self):
        wraps = np.zeros(self.r.shape, dtype=np.int)
        for i_dim in np.where(np.isfinite(self.L))[0]:
            wraps_mag = ((np.abs(self.r[:, i_dim]) + self.L[i_dim] / 2.0) //
                         self.L[i_dim])
            wraps[:, i_dim] = np.sign(self.r[:, i_dim]) * wraps_mag
        return wraps

    def __repr__(self):
        fs = [('n', self.n), ('dim', self.dim), ('L', self.L)]
        return make_repr_str(self, fs)


class NonePositions(object):

    def __repr__(self):
        fs = []
        return make_repr_str(self, fs)


def positions_factory(spatial_flag, periodic_flag, n, dim=None,
                      L=None, origin_flags=None,
                      rng=None, obstructor=None):
    if not spatial_flag:
        return NonePositions()
    elif not periodic_flag:
        r_0 = np.zeros([n, dim])
        return Positions(r_0)
    else:
        r_0 = get_uniform_points(n, L, origin_flags, rng, obstructor)
        return PeriodicPositions(L, r_0)


def get_uniform_points(n, L, origin_flags=None, rng=None, obstructor=None):
    dim = L.shape[0]
    if origin_flags is None:
        origin_flags = np.zeros([dim], dtype=np.bool)
    if rng is None:
        rng = np.random
    r = np.zeros([n, dim])
    for i_n in range(n):
        while True:
            for i_dim in range(dim):
                if np.isinf(L[i_dim]) or origin_flags[i_dim]:
                    r[i_n, i_dim] = 0.0
                else:
                    r[i_n, i_dim] = rng.uniform(-L[i_dim] / 2.0,
                                                L[i_dim] / 2.0)
            if obstructor is None:
                break
            elif not obstructor.get_obstructeds(np.array([r[i_n]]))[0]:
                break
    return r



from __future__ import print_function, division
import numpy as np
from ciabatta import crandom
from ciabatta.meta import make_repr_str
from spatious import vector
from spatious.vector import smallest_signed_angle as angle_dist


class Turner(object):

    def get_angle(self, th_in, *args, **kwargs):
        return th_in

    def get_norm_angle(self, *args, **kwargs):
        return vector.normalise_angle(self.get_angle(*args, **kwargs))

    def turn(self, obs, ds, *args, **kwargs):
        ds.th[obs] = self.get_norm_angle(ds.th[obs], *args, **kwargs)

    def __repr__(self):
        fs = []
        return make_repr_str(self, fs)


class BounceBackTurner(Turner):

    def get_angle(self, th_in, *args, **kwargs):
        return th_in + np.pi


class ReflectTurner(Turner):

    def get_angle(self, th_in, th_normal, *args, **kwargs):
        th_rel = th_in - th_normal
        return th_normal + np.where(th_rel == 0.0,
                                    np.pi,
                                    np.sign(th_rel) * np.pi - th_rel)


class AlignTurner(Turner):

    def get_angle(self, th_in, th_normal, rng, *args, **kwargs):
        if rng is None:
            rng = np.random
        th_rel = vector.normalise_angle(th_in - th_normal)
        antiparallels = np.isclose(np.abs(angle_dist(th_in, th_normal)), np.pi)
        signs = np.where(antiparallels,
                         crandom.randbool(antiparallels.shape[0], rng),
                         np.sign(th_rel))
        th_rel = signs * np.pi / 2.0
        return th_normal + th_rel



from __future__ import print_function, division
import numpy as np
from ciabatta.meta import make_repr_str
import ahoy
from ahoy import ships, obstructors, fields, turners


class Model(object):

    def __init__(self, seed, dt,
                 aligned_flag=False, origin_flags=None,
                 **ship_kwargs):
        # Record initial conditions
        self.seed = seed
        self.dt = dt

        # Initialise system
        self.i = 0
        self.rng = np.random.RandomState(seed)
        self.ships = ships.ships_factory(self.rng,
                                         aligned_flag=aligned_flag,
                                         origin_flags=origin_flags,
                                         **ship_kwargs)

    @property
    def aligned_flags(self):
        return np.all(self.ships.agents.directions.u_0 == 0.0, axis=0)

    @property
    def origin_flags(self):
        return np.all(self.ships.agents.positions.r_0 == 0.0, axis=0)

    def iterate(self):
        self.ships.iterate(self.dt, self.rng)
        self.i += 1

    def _get_output_dirname_agent_part(self):
        ags = self.ships.agents

        origin_str = '({})'.format(','.join(['{:d}'.format(a)
                                             for a in self.aligned_flags]))
        s = 'n={},align={}'.format(ags.positions.n, origin_str)

        # Space and swimming
        if isinstance(ags.positions, ahoy.positions.Positions):
            origin_str = '({})'.format(','.join(['{:d}'.format(o)
                                                 for o in self.origin_flags]))
            s += ',origin={},v={:g}'.format(origin_str, ags.swimmers.v_0)
            if isinstance(ags.positions, ahoy.positions.PeriodicPositions):
                def format_inf(x):
                    return '{:g}'.format(x) if np.isfinite(x) else 'i'
                L_str = '({})'.format(','.join([format_inf(e)
                                                for e in ags.positions.L]))
                s += ',L={}'.format(L_str)

        # Rudders
        for rs in ags.rudder_sets.sets:
            nm = rs.noise_measurer
            if isinstance(rs, ahoy.rudders.TumbleRudders):
                noise_str = 'p'
            elif isinstance(rs, ahoy.rudders.RotationRudders):
                noise_str = 'Dr'
            s += ',{}={:g}'.format(noise_str, nm.noise_0)
            if rs.is_chemotactic:
                type_s = 'T' if nm.is_temporal else 'S'
                side = 2 - rs.is_onesided
                s += ',chi={:.2g},side={:d},type={}'.format(nm.chi, side,
                                                           type_s)
                if nm.is_temporal:
                    measurer = nm.dc_dx_measurer
                    s += ',dtMem={:g},tMem={:g}'.format(measurer.dt_mem,
                                                        measurer.t_mem)
        return s

    def _get_output_dirname_obstruction_part(self):
        obs = self.ships.obstructor
        s = 'obs='

        if obs.__class__ is obstructors.NoneObstructor:
            s += 'NoObs'
        elif obs.__class__ is obstructors.PorousObstructor:
            if obs.turner.__class__ is turners.Turner:
                s_turner = 'stall'
            elif obs.turner.__class__ is turners.BounceBackTurner:
                s_turner = 'bback'
            elif obs.turner.__class__ is turners.ReflectTurner:
                s_turner = 'reflect'
            elif obs.turner.__class__ is turners.AlignTurner:
                s_turner = 'align'
            pf = obs.fraction_occupied
            s += 'Pore(R={:g},pf={:.2g},turn={})'.format(obs.R, pf, s_turner)
        return s

    def _get_output_dirname_field_part(self):
        c_field = self.ships.c_field
        s = 'c='
        if c_field.__class__ is fields.NoneFoodField:
            s += 'NoC'
        elif c_field.__class__ is fields.FoodField:
            s += 'C(c0={:g},cD={:g},cDelta={:g})'.format(c_field.c_0,
                                                         c_field.D,
                                                         c_field.delta)
        return s

    def get_output_dirname(self):
        s = 'ahoy_{}D,dt={:g},seed={}'.format(self.ships.dim, self.dt,
                                              self.seed)
        s += ',{}'.format(self._get_output_dirname_agent_part())
        s += ',{}'.format(self._get_output_dirname_obstruction_part())
        s += ',{}'.format(self._get_output_dirname_field_part())
        return s

    def __repr__(self):
        fs = [('seed', self.seed), ('dt', self.dt),
              ('origin_flags', self.origin_flags),
              ('aligned_flags', self.aligned_flags),
              ('i', self.i), ('rng', self.rng),
              ('ships', self.ships)
              ]
        return make_repr_str(self, fs)



import numpy as np
from ahoy import obstructors, turners
import test


class TestPorousObstructor(test.TestBase):

    def test_random_seeding(self):
        obstructor_kwargs = {
            'turner': turners.ReflectTurner(),
            'R': 50.0,
            'L': np.array([200.0, 200.0]),
            'pf': 0.4,
            'periodic_flag': True,
        }

        rng_seed = 1

        def get_obstructor(npy_seed):
            np.random.seed(npy_seed)
            rng = np.random.RandomState(rng_seed)
            obstructor = obstructors.PorousObstructor(rng=rng,
                                                      **obstructor_kwargs)
            return obstructor

        obstructor_1 = get_obstructor(2)
        obstructor_2 = get_obstructor(3)

        self.assertTrue(np.allclose(obstructor_1.R, obstructor_2.R))
        self.assertTrue(np.allclose(obstructor_1.L, obstructor_2.L))
        self.assertTrue(np.allclose(obstructor_1.rs, obstructor_2.rs))



from __future__ import print_function, division
from abc import ABCMeta, abstractmethod
import numpy as np
from ciabatta.meta import make_repr_str
from ahoy.ring_buffer import CylinderBuffer
from ahoy import measurers, c_measurers


def get_K(t, dt, t_rot_0):
    A = 0.5
    ts = np.arange(0.0, t, dt)
    gs = ts / t_rot_0
    K = np.exp(-gs) * (1.0 - A * (gs + (gs ** 2) / 2.0))
    K[K < 0.0] *= np.abs(K[K >= 0.0].sum() / K[K < 0.0].sum())
    K /= np.sum(K * -ts * dt)
    return K


class DcDxMeasurer(measurers.Measurer):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_dc_dxs(self):
        return


class SpatialDcDxMeasurer(DcDxMeasurer):

    def __init__(self, directions, grad_c_measurer):
        self.directions = directions
        self.grad_c_measurer = grad_c_measurer

    def get_dc_dxs(self):
        grad_c = self.grad_c_measurer.get_grad_cs()
        return np.sum(self.directions.u * grad_c, axis=-1)

    def __repr__(self):
        fs = [('grad_c_measurer', self.grad_c_measurer)]
        return make_repr_str(self, fs)


class TemporalDcDxMeasurer(DcDxMeasurer):

    def __init__(self, c_measurer, v_0, dt_mem, t_mem, t_rot_0,
                 time):
        self.c_measurer = c_measurer
        self.v_0 = v_0
        self.dt_mem = dt_mem
        self.t_mem = t_mem
        cs = self.c_measurer.get_cs()
        n = cs.shape[0]
        self.K_dt = get_K(self.t_mem, self.dt_mem, t_rot_0) * self.dt_mem
        self.c_mem = CylinderBuffer(n, self.K_dt.shape[0])
        self.time = time

        # Optimisation, only calculate dc_dx when c memory is updated.
        self.dc_dx_cache = np.zeros([n])
        self.t_last_update = 0.0

    def _iterate(self):
        cs = self.c_measurer.get_cs()
        self.c_mem.update(cs)

    def _get_dc_dxs(self):
        return self.c_mem.integral_transform(self.K_dt) / self.v_0

    def iterate(self):
        t_now = self.time.t
        if t_now - self.t_last_update > 0.99 * self.dt_mem:
            self._iterate()
            self.dc_dx_cache = self._get_dc_dxs()
            self.t_last_update = t_now

    def get_dc_dxs(self):
        self.iterate()
        return self.dc_dx_cache

    def __repr__(self):
        fs = [('c_measurer', self.c_measurer), ('v_0', self.v_0),
              ('dt_mem', self.dt_mem), ('t_mem', self.t_mem),
              ('t_last_update', self.t_last_update)]
        return make_repr_str(self, fs)


def dc_dx_factory(temporal_chemo_flag,
                  ds=None,
                  ps=None, v_0=None, dt_mem=None, t_mem=None, t_rot_0=None, time=None,
                  c_field_flag=None, c_field=None):
    if temporal_chemo_flag:
        return temporal_dc_dx_factory(ps, v_0, dt_mem, t_mem, t_rot_0, time,
                                      c_field_flag, c_field)
    else:
        return spatial_dc_dx_factory(ds, c_field_flag, c_field, ps)


def spatial_dc_dx_factory(ds, c_field_flag=None, c_field=None, ps=None):
    if not c_field_flag:
        grad_c_measurer = c_measurers.ConstantGradCMeasurer(ds.n, ds.dim)
    else:
        grad_c_measurer = c_measurers.FieldGradCMeasurer(c_field, ps)
    return SpatialDcDxMeasurer(ds, grad_c_measurer)


def temporal_dc_dx_factory(ps, v_0, dt_mem, t_mem, t_rot_0, time,
                           c_field_flag=None, c_field=None):
    if not c_field_flag:
        c_measurer = c_measurers.LinearCMeasurer(ps)
    else:
        c_measurer = c_measurers.FieldCMeasurer(c_field, ps)
    return TemporalDcDxMeasurer(c_measurer, v_0, dt_mem, t_mem, t_rot_0, time)



from __future__ import print_function, division
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from ciabatta.ejm_rcparams import reds_cmap
from agaro import output_utils
from ahoy.utils import utils
from ahoy.plot.var_plot import VarPlot
import ahoy.fields


def plot_2d(dirname):
    fig = plt.figure()
    ax_vis = fig.add_subplot(111)

    fnames = output_utils.get_filenames(dirname)
    m_0 = output_utils.filename_to_model(fnames[0])

    L = m_0.ships.agents.positions.L

    ax_vis.set_xlim(-L[0] / 2.0, L[0] / 2.0)
    ax_vis.set_ylim(-L[1] / 2.0, L[1] / 2.0)
    ax_vis.set_aspect('equal')

    plt.subplots_adjust(left=0.25, bottom=0.25)
    has_c_field = isinstance(m_0.ships.c_field, ahoy.fields.FoodField)
    if has_c_field:
        plot_c = VarPlot(m_0.ships.c_field.c, cmap=reds_cmap, axes=ax_vis)
    plot_p = ax_vis.quiver(m_0.ships.agents.positions.r_w[:, 0],
                           m_0.ships.agents.positions.r_w[:, 1],
                           m_0.ships.agents.directions.u[:, 0],
                           m_0.ships.agents.directions.u[:, 1])

    ax_slide = plt.axes([0.25, 0.1, 0.65, 0.03])
    t_slider = Slider(ax_slide, 'Time', 0, len(fnames), valinit=0)

    t_time = fig.text(0.1, 0.5, '')

    def update(val):
        fname_i = int(round(val))
        if 0 <= fname_i < len(fnames):
            m = output_utils.filename_to_model(fnames[fname_i])
            if has_c_field:
                plot_c.update(m.ships.c_field.c)
            plot_p.set_offsets(m.ships.agents.positions.r_w)
            plot_p.set_UVC(m.ships.agents.directions.u[:, 0],
                           m.ships.agents.directions.u[:, 1])
            t_time.set_text('Time: {:g}'.format(m.ships.time.t))

            fig.canvas.draw_idle()

    t_slider.on_changed(update)

    plt.show()


def plot_linear_density(dirname):
    fig = plt.figure()
    ax_d = fig.add_subplot(211)
    ax_c = fig.add_subplot(212)

    fnames = output_utils.get_filenames(dirname)
    m_0 = output_utils.filename_to_model(fnames[0])

    L = m_0.ships.agents.positions.L

    dx = L[0] / 100.0

    plt.subplots_adjust(left=0.25, bottom=0.25)

    ds, xbs = utils.get_linear_density(m_0, dx)

    plot_d = ax_d.bar(xbs[:-1], ds, width=xbs[1] - xbs[0])
    c_field = m_0.ships.c_field.c
    plot_c = ax_c.scatter(c_field.mesh.cellCenters[0, :], c_field.value)
    ax_slide = plt.axes([0.25, 0.1, 0.65, 0.03])
    t_slider = Slider(ax_slide, 'Index', 0, len(fnames), valinit=0)

    ax_d.set_xlim(-L[0] / 2.0, L[0] / 2.0)
    ax_c.set_xlim(-L[0] / 2.0, L[0] / 2.0)
    ax_c.set_ylim(0.0, m_0.ships.c_field.c_0)

    def update(val):
        fname_i = int(round(val))
        if 0 <= fname_i < len(fnames):
            m = output_utils.filename_to_model(fnames[fname_i])
            ds, xbs = utils.get_linear_density(m, dx)
            for rect, d in zip(plot_d, ds):
                rect.set_height(d)
            c_field = m.ships.c_field.c
            plot_c.set_offsets(np.array([c_field.mesh.cellCenters[0, :],
                                         c_field.value]).T)
            fig.canvas.draw_idle()

    t_slider.on_changed(update)

    plt.show()


def plot_1d(dirname):
    fig = plt.figure()
    ax_vis = fig.add_subplot(211)
    ax_d = fig.add_subplot(212)

    fnames = output_utils.get_filenames(dirname)
    m_0 = output_utils.filename_to_model(fnames[0])

    L = m_0.ships.agents.positions.L

    ax_vis.set_xlim(-L[0] / 2.0, L[0] / 2.0)
    ax_d.set_xlim(-L[0] / 2.0, L[0] / 2.0)

    dx = L / 100.0

    plt.subplots_adjust(left=0.25, bottom=0.25)
    plot_p = ax_vis.scatter(m_0.ships.agents.positions.r_w[:, 0],
                            np.zeros([m_0.ships.agents.n]))

    d = m_0.ships.agents.positions.get_density_field(dx)
    x = np.linspace(-L[0] / 2.0, L[0] / 2.0, d.shape[0])

    plot_d = ax_d.bar(x, d, width=x[1] - x[0])

    ax_slide = plt.axes([0.25, 0.1, 0.65, 0.03])
    t_slider = Slider(ax_slide, 'Index', 0, len(fnames), valinit=0)

    def update(val):
        fname_i = int(round(val))
        if 0 <= fname_i < len(fnames):
            m = output_utils.filename_to_model(fnames[fname_i])
            plot_p.set_offsets(np.array([m.ships.agents.positions.r_w[:, 0],
                                         np.zeros([m.ships.agents.n])]).T)
            ds = m.ships.agents.positions.get_density_field(dx)
            for rect, d in zip(plot_d, ds):
                rect.set_height(d)
            ax_d.set_ylim(0.0, 1.05 * ds.max())
            fig.canvas.draw_idle()

    t_slider.on_changed(update)

    plt.show()


def plot_vis(dirname):
    dim = output_utils.get_recent_model(dirname).ships.dim
    if dim == 1:
        plot_1d(dirname)
    elif dim == 2:
        plot_2d(dirname)


def plot_t_uds_scalar(dirname, ax):
    ts, uds = utils.t_uds_scalar(dirname)
    ax.plot(ts, uds)


def plot_t_uds_vector(dirname, ax):
    ts, uds = utils.t_uds_vector(dirname)
    for vd_set in uds.T:
        ax.plot(ts, vd_set)


def plot_t_Ds_scalar(dirname, ax):
    ts, Ds = utils.t_Ds_scalar(dirname)
    ax.plot(ts, Ds)


def plot_t_Ds_vector(dirname, ax):
    ts, Ds = utils.t_Ds_vector(dirname)
    for D_set in Ds.T:
        ax.plot(ts, D_set)


def plot_t_rs_scalar(dirname, ax):
    ts, rs = utils.t_rs_scalar(dirname)
    ax.plot(ts, rs)


def plot_t_rs_vector(dirname, ax):
    ts, rs = utils.t_rs_vector(dirname)
    for r_set in rs.T:
        ax.plot(ts, r_set)


def plot_t_u_nets_scalar(dirname, ax):
    ts, u_nets = utils.t_u_nets_scalar(dirname)
    ax.plot(ts, u_nets)


def plot_t_u_nets_vector(dirname, ax):
    ts, u_nets = utils.t_u_nets_vector(dirname)
    for u_net_set in u_nets.T:
        ax.plot(ts, u_net_set)


def plot_chi_uds_x(dirnames, ax):
    chis, uds = utils.chi_uds_x(dirnames)
    i_sort = np.argsort(chis)
    chis, uds = chis[i_sort], uds[i_sort]
    ax.scatter(chis, uds)
    ax.set_xlim(-0.02, 1.0)
    ax.set_ylim(0.0, 1.1)


def plot_pf_Ds_scalar(dirnames, ax):
    pfs, Ds = utils.pf_Ds_scalar(dirnames)
    i_sort = np.argsort(pfs)
    pfs, Ds = pfs[i_sort], Ds[i_sort]
    ax.scatter(pfs, Ds)
    ax.set_xlim(-0.02, 1.0)
    ax.set_ylim(0.0, 410.0)


def plot_pf_uds_x(dirnames, ax):
    pfs, uds = utils.pf_uds_x(dirnames)
    i_sort = np.argsort(pfs)
    pfs, uds = pfs[i_sort], uds[i_sort]
    ax.scatter(pfs, uds / uds[0])
    ax.set_xlim(-0.02, 1.0)
    ax.set_ylim(0.0, 1.01)


def plot_pf_duds_x(dirnames, ax):
    pfs, uds = utils.pf_uds_x(dirnames)
    i_sort = np.argsort(pfs)
    pfs, uds = pfs[i_sort], uds[i_sort]
    uds_norm = uds / uds[0]
    duds = np.diff(uds_norm) / pfs[:-1]
    ax.scatter(pfs[:-1], duds)
    ax.set_xlim(-0.02, 1.0)
    # ax.set_ylim(0.0, 1.0)


def plot_Dr_0_Ds_scalar(dirnames, ax):
    Dr_0s, Ds = utils.Dr_0_Ds_scalar(dirnames)
    i_sort = np.argsort(Dr_0s)
    Dr_0s, Ds = Dr_0s[i_sort], Ds[i_sort]
    ax.scatter(Dr_0s, Ds)
    # ax.set_xlim(-0.02, 1.0)
    # ax.set_ylim(0.0, 410.0)
    ax.set_xscale('log')
    ax.set_yscale('log')


def plot_p_0_Ds_scalar(dirnames, ax):
    p_0s, Ds = utils.p_0_Ds_scalar(dirnames)
    i_sort = np.argsort(p_0s)
    p_0s, Ds = p_0s[i_sort], Ds[i_sort]
    ax.scatter(p_0s, Ds)
    # ax.set_xlim(-0.02, 1.0)
    # ax.set_ylim(0.0, 410.0)
    ax.set_xscale('log')
    ax.set_yscale('log')


def plot_th_density(ds, th_bins, fig):
    ax = plt.subplot(111, polar=True)
    dth = th_bins[1] - th_bins[0]
    ax.plot(th_bins[:-1], ds * dth)
    ax.set_rmax(1.0)
    ax.grid(True)



from __future__ import print_function, division
import numpy as np
from ahoy import positions
import test


class TestPeriodicPositions1D(test.TestBase):
    dim = 1
    L = np.array([1.7])

    def setUp(self):
        super(TestPeriodicPositions1D, self).setUp()
        self.ps = positions.positions_factory(spatial_flag=True,
                                              periodic_flag=True,
                                              n=self.n, L=self.L,
                                              rng=self.rng)

    def test_wrapping_down(self):
        dr = np.zeros_like(self.ps.r)
        self.ps.r[0, 0] = self.L[0] / 2.0
        dr[0, 0] = 0.9 * self.L[0]
        self.ps.r += np.full(self.ps.r.shape, dr)
        self.assertTrue(np.abs(self.ps.r_w[:, 0]).max() < self.L[0] / 2.0)

    def test_wrapping_up(self):
        dr = np.zeros_like(self.ps.r)
        self.ps.r[-1, 0] = -self.L[0] / 2.0
        dr[-1, 0] = -0.9 * self.L[0]
        self.ps.r += np.full(self.ps.r.shape, dr)
        self.assertTrue(np.abs(self.ps.r_w[:, 0]).max() < self.L[0] / 2.0)


class TestPeriodicPositions2D(TestPeriodicPositions1D):
    dim = 2
    L = np.array([0.5, 1.5])

    def test_wrapping_up(self):
        dr = np.zeros_like(self.ps.r)
        self.ps.r[-1, -1] = -self.L[-1] / 2.0
        dr[-1, -1] = -0.9 * self.L[-1]
        self.ps.r += np.full(self.ps.r.shape, dr)
        self.assertTrue(np.abs(self.ps.r_w[:, -1]).max() < self.L[-1] / 2.0)

    def test_partial_infinite_boundaries(self):
        L = np.array([np.inf, 1.7])
        ps = positions.positions_factory(spatial_flag=True,
                                         periodic_flag=True,
                                         n=self.n, L=L, rng=self.rng)
        dr = self.rng.uniform(-1.0, 1.0, size=ps.r.shape)
        ps.r += dr
        r_naive = ps.r_0[:, 0] + dr[:, 0]
        # Check no wrapping along infinite axis
        self.assertTrue(np.allclose(r_naive, ps.r_w[:, 0]))
        self.assertTrue(np.allclose(r_naive, ps.r[:, 0]))
        # Check done wrapping along finite axis
        self.assertTrue(np.all(np.abs(ps.r_w[:, 1]) < L[1] / 2.0))



from __future__ import print_function, division
import numpy as np
from ahoy import turners
from ciabatta.vector import smallest_signed_angle as angle_dist
import test


class TestTurner(test.TestBase):
    turner_cls = turners.Turner
    left = np.pi
    right = 0.0
    up = np.pi / 2.0
    down = -np.pi / 2.0

    NE = np.pi / 4.0
    NW = 3.0 * np.pi / 4.0
    SW = -3.0 * np.pi / 4.0
    SE = -np.pi / 4.0

    def setUp(self):
        super(TestTurner, self).setUp()
        self.turner = self.turner_cls()

    def do_turning(self, th_normal, th_in, th_out_expected):
        print('Normal: {:g}'.format(th_normal))
        print('In: {:g}'.format(th_in))
        th_in = np.array([th_in])
        th_normal = np.array([th_normal])
        th_out = self.turner.get_norm_angle(th_in, th_normal, self.rng)
        print('Out: {:g}'.format(th_out[0]))
        print('Expected: {:g}'.format(th_out_expected))
        self.assertTrue(np.allclose(angle_dist(th_out, th_out_expected), 0.0))

    def test_right_left(self):
        self.do_turning(self.left, self.right, self.right)

    def test_up_left(self):
        self.do_turning(self.left, self.up, self.up)

    def test_down_left(self):
        self.do_turning(self.left, self.down, self.down)

    def test_equivalent_angles(self):
        th_in = self.NE
        th_in = np.array([th_in])
        th_outs = []
        for i in range(-2, 2):
            th_normal = np.pi + i * 2.0 * np.pi
            th_normal = np.array([th_normal])
            th_out = self.turner.get_norm_angle(th_in, th_normal, self.rng)
            th_outs.append(th_out[0])
        print(th_outs)
        dths = angle_dist(th_outs, th_outs[0])
        print(dths)
        self.assertTrue(np.allclose(dths, 0.0))


class TestBounceBackTurner(TestTurner):
    turner_cls = turners.BounceBackTurner

    def test_right_left(self):
        self.do_turning(self.left, self.right, self.left)

    def test_up_left(self):
        self.do_turning(self.left, self.up, self.down)

    def test_down_left(self):
        self.do_turning(self.left, self.down, self.up)


class TestReflectTurner(TestTurner):
    turner_cls = turners.ReflectTurner

    def test_right_left(self):
        self.do_turning(self.left, self.right, self.left)

    def test_up_left(self):
        self.do_turning(self.left, self.up, self.up)

    def test_down_left(self):
        self.do_turning(self.left, self.down, self.down)

    def test_left_right(self):
        self.do_turning(self.right, self.left, self.right)

    def test_up_right(self):
        self.do_turning(self.right, self.up, self.up)

    def test_down_right(self):
        self.do_turning(self.right, self.down, self.down)

    def test_right_up(self):
        self.do_turning(self.up, self.right, self.right)

    def test_left_up(self):
        self.do_turning(self.up, self.left, self.left)

    def test_down_up(self):
        self.do_turning(self.up, self.down, self.up)

    def test_right_down(self):
        self.do_turning(self.down, self.right, self.right)

    def test_left_down(self):
        self.do_turning(self.down, self.left, self.left)

    def test_up_down(self):
        self.do_turning(self.down, self.up, self.down)


class TestAlignTurner(TestTurner):
    turner_cls = turners.AlignTurner

    def test_up_left(self):
        self.do_turning(self.left, self.up, self.up)

    def test_down_left(self):
        self.do_turning(self.left, self.down, self.down)

    def test_up_right(self):
        self.do_turning(self.right, self.up, self.up)

    def test_down_right(self):
        self.do_turning(self.right, self.down, self.down)

    def test_right_up(self):
        self.do_turning(self.up, self.right, self.right)

    def test_left_up(self):
        self.do_turning(self.up, self.left, self.left)

    def test_right_down(self):
        self.do_turning(self.down, self.right, self.right)

    def test_left_down(self):
        self.do_turning(self.down, self.left, self.left)

    def test_NE_left(self):
        self.do_turning(self.left, self.NE, self.up)

    def test_SE_left(self):
        self.do_turning(self.left, self.SE, self.down)

    def test_NW_right(self):
        self.do_turning(self.right, self.NW, self.up)

    def test_SW_right(self):
        self.do_turning(self.right, self.SW, self.down)

    def do_antiparallel(self, th_in, th_normal, th_out_1, th_out_2):
        self.turner = self.turner_cls()
        n = 1000
        th_in = np.full([n], th_in)
        th_normal = np.full([n], th_normal)

        th_out = self.turner.get_norm_angle(th_in, th_normal, self.rng)
        print(th_out)

        frac_1 = np.isclose(angle_dist(th_out, th_out_1), 0.0).sum() / float(n)
        self.assertAlmostEqual(frac_1, 0.5, 1)
        frac_2 = np.isclose(angle_dist(th_out, th_out_2), 0.0).sum() / float(n)
        self.assertAlmostEqual(frac_2, 0.5, 1)

    def test_right_left(self):
        self.do_antiparallel(self.right, self.left, self.up, self.down)

    def test_down_up(self):
        self.do_antiparallel(self.down, self.up, self.left, self.right)

    def test_NE_SW(self):
        self.do_antiparallel(self.NE, self.SW, self.NW, self.SE)



import fipy


gmsh_text_box = '''
// Define the square that acts as the system boundary.

dx = %(dx)g;
Lx = %(Lx)g;
Ly = %(Ly)g;
p_n_w = newp; Point(p_n_w) = {-Lx / 2.0, Ly / 2.0, 0, dx};
p_n_e = newp; Point(p_n_e) = {Lx / 2.0, Ly / 2.0, 0, dx};
p_s_e = newp; Point(p_s_e) = {Lx / 2.0, -Ly / 2.0, 0, dx};
p_s_w = newp; Point(p_s_w) = {-Lx / 2.0, -Ly / 2.0, 0, dx};
l_n = newl; Line(l_n) = {p_n_w, p_n_e};
l_e = newl; Line(l_e) = {p_n_e, p_s_e};
l_s = newl; Line(l_s) = {p_s_e, p_s_w};
l_w = newl; Line(l_w) = {p_s_w, p_n_w};
ll = newll; Line Loop(ll) = {l_n, l_e, l_s, l_w};

'''

gmsh_text_circle = '''
// Define a circle that acts as an obstacle
x = %(x)g;
y = %(y)g;
R = %(R)g;
p_c = newp; Point(p_c) = {x, y, 0, dx};
p_w = newp; Point(p_w) = {x - R, y, 0, dx};
p_n = newp; Point(p_n) = {x, y + R, 0, dx};
p_e = newp; Point(p_e) = {x + R, y, 0, dx};
p_s = newp; Point(p_s) = {x, y - R, 0, dx};
c_w_n = newreg; Circle(c_w_n) = {p_w, p_c, p_n};
c_n_e = newreg; Circle(c_n_e) = {p_n, p_c, p_e};
c_e_s = newreg; Circle(c_e_s) = {p_e, p_c, p_s};
c_s_w = newreg; Circle(c_s_w) = {p_s, p_c, p_w};
Line Loop(%(i)d) = {c_w_n, c_n_e, c_e_s, c_s_w};

'''

gmsh_text_surface = '''
// The first argument is the outer loop boundary.
// The remainder are holes in it.
Plane Surface(1) = {ll, %(args)s};
'''


def _porous_mesh_geo_factory(rs, R, dx, L):
    gmsh_text = gmsh_text_box % {'dx': dx, 'Lx': L[0], 'Ly': L[1]}
    circle_loop_indexes = []
    if rs is not None and len(rs) and R:
        for i in range(len(rs)):
            index = 10 * (i + 1)
            gmsh_text += gmsh_text_circle % {'x': rs[i][0], 'y': rs[i][1],
                                             'R': R, 'i': index}
            circle_loop_indexes += [index]
    surface_args = ', '.join([str(i) for i in circle_loop_indexes])
    gmsh_text += gmsh_text_surface % {'args': surface_args}
    return gmsh_text


def uniform_mesh_factory(L, dx):
    dim = len(L)
    if dim == 1:
        return fipy.Grid1D(dx=dx, Lx=L[0]) - L[0] / 2.0
    elif dim == 2:
        return (fipy.Grid2D(dx=dx, dy=dx, Lx=L[0], Ly=L[1]) -
                ((L[0] / 2.0,), (L[1] / 2.0,)))


def porous_mesh_factory(rs, R, dx, L):
    return fipy.Gmsh2D(_porous_mesh_geo_factory(rs, R, dx, L))



import numpy as np
from ahoy.model import Model
import test


class TestModel(test.TestBase):

    def test_model_random_seeding(self):
        model_kwargs = {
            'seed': 1,
            'dt': 0.01,

            'n': 10,
            'dim': 2,
            # Must have aligned flag False to test uniform directions function.
            'aligned_flag': False,

            'spatial_flag': False,

            'chi': 0.1,
            'onesided_flag': True,

            # Must have tumbling to test tumbling function.
            'tumble_flag': True,
            'p_0': 1.3,
            'tumble_chemo_flag': True,

            # Must have rotational diffusion to test rot diff function.
            'rotation_flag': True,
            'Dr_0': 1.3,
            'rotation_chemo_flag': True,
        }

        num_iterations = 100

        def get_model(npy_seed):
            np.random.seed(npy_seed)
            model = Model(**model_kwargs)
            for _ in range(num_iterations):
                model.iterate()
            return model

        model_1 = get_model(2)
        model_2 = get_model(3)

        self.assertTrue(np.allclose(model_1.ships.agents.directions.u,
                                    model_2.ships.agents.directions.u))

    def test_spatial_model_random_seeding(self):
        model_kwargs = {
            'seed': 1,
            'dt': 0.01,

            'dim': 2,
            'n': 10,
            # Must have aligned flag False to test uniform directions function.
            'aligned_flag': False,

            'spatial_flag': True,
            'periodic_flag': True,
            'v_0': 1.5,
            # Must have at least one periodic axis to test uniform points
            # function.
            'L': np.array([2.0, 2.2]),
            'origin_flags': np.array([False, False]),

            'chi': 0.1,
            'onesided_flag': True,

            # Must have tumbling to test tumbling function.
            'tumble_flag': True,
            'p_0': 1.3,
            'tumble_chemo_flag': True,

            # Must have rotational diffusion to test rot diff function.
            'rotation_flag': True,
            'Dr_0': 1.3,
            'rotation_chemo_flag': True,

            'temporal_chemo_flag': True,
            'dt_mem': 0.05,
            't_mem': 5.0,

            'pore_flag': True,
            'pore_turner': 'align',
            'pore_R': 0.1,
            'pore_pf': 0.1,
        }

        num_iterations = 100

        def get_model(npy_seed):
            np.random.seed(npy_seed)
            model = Model(**model_kwargs)
            for _ in range(num_iterations):
                model.iterate()
            return model

        model_1 = get_model(2)
        model_2 = get_model(3)

        self.assertTrue(np.allclose(model_1.ships.agents.positions.r,
                                    model_2.ships.agents.positions.r))
        self.assertTrue(np.allclose(model_1.ships.agents.directions.u,
                                    model_2.ships.agents.directions.u))

    def test_c_field_model_random_seeding(self):
        model_kwargs = {
            'seed': 1,
            'dt': 0.01,

            'dim': 2,
            'rho_0': 10.0,
            # Must have aligned flag False to test uniform directions function.
            'aligned_flag': False,

            'spatial_flag': True,
            'periodic_flag': True,
            'v_0': 1.5,
            'L': np.array([2.0, 2.2]),
            'origin_flags': np.array([False, False]),

            'chi': 0.3,
            'onesided_flag': True,

            # Must have tumbling to test tumbling function.
            'p_0': 1.3,
            'tumble_flag': True,
            'tumble_chemo_flag': True,

            # Must have rotational diffusion to test rot diff function.
            'rotation_flag': True,
            'Dr_0': 1.3,
            'rotation_chemo_flag': True,

            'temporal_chemo_flag': True,
            'dt_mem': 0.1,
            't_mem': 5.0,

            'pore_flag': True,
            'pore_turner': 'align',
            'pore_R': 0.2,
            'pore_pf': 0.1,

            'c_dx': 0.2,
            'c_D': 10.0,
            'c_delta': 1000.0,
            'c_0': 1.3,
        }

        num_iterations = 100

        def get_model(npy_seed):
            np.random.seed(npy_seed)
            model = Model(**model_kwargs)
            for _ in range(num_iterations):
                model.iterate()
            return model

        model_1 = get_model(2)
        model_2 = get_model(3)

        self.assertTrue(np.allclose(model_1.ships.agents.positions.r,
                                    model_2.ships.agents.positions.r))
        self.assertTrue(np.allclose(model_1.ships.agents.directions.u,
                                    model_2.ships.agents.directions.u))



from __future__ import print_function, division
from itertools import product
import numpy as np
from scipy.stats import multivariate_normal
from ahoy import positions, fields
from ahoy.mesh import uniform_mesh_factory
import test


def get_nearest_cell_ids_manual(f, ps):
    rs = ps.r_w.T
    ccs = f.mesh.cellCenters.value
    drs = np.abs(ccs[:, np.newaxis, :] - rs[:, :, np.newaxis])
    for i_dim in range(ps.dim):
        drs[i_dim] = np.minimum(drs[i_dim], ps.L[i_dim] - drs[i_dim])
    dr_mags = np.sum(np.square(drs), axis=0)
    return np.argmin(dr_mags, axis=1)


class TestField1D(test.TestBase):
    L = np.array([1.6])
    dx = 0.1

    x_vals = [-L[0] / 2.012, -0.312, 0.01, 0.121, L[0] / 1.976]
    rs_special = np.array(list(product(x_vals)))

    def test_field(self):
        n = 100

        mesh = uniform_mesh_factory(self.L, self.dx)
        f = fields.Field(mesh, c_0=1.0)

        rs_random = positions.get_uniform_points(n, self.L, rng=self.rng)
        rs = np.append(rs_random, self.rs_special, axis=0)

        ps = positions.PeriodicPositions(self.L, rs)
        cids = f.get_nearest_cell_ids(ps)
        cids_manual = get_nearest_cell_ids_manual(f, ps)
        self.assertTrue(np.allclose(cids, cids_manual))


class TestField2D(TestField1D):
    L = np.array([1.0, 2.0])

    x_vals = [-L[0] / 2.012, -0.312, 0.01, 0.121, L[0] / 1.976]
    y_vals = [-L[1] / 1.99632, -0.312, 0.01, 0.121, L[1] / 2.0021]
    rs_special = np.array(list(product(x_vals, y_vals)))


class TestFoodField1D(test.TestBase):
    L = np.array([4.0])
    dx = 0.005

    def test_rho_array_uniform(self):
        dim = self.L.shape[0]
        rho_expected = 1.0 / (self.dx ** dim)

        mesh = uniform_mesh_factory(self.L, self.dx)
        r_centers = mesh.cellCenters.value.T
        ps_centers = positions.PeriodicPositions(self.L, r_centers)

        D = 1.0
        delta = 1.0
        c_0 = 1.0
        f = fields.FoodField(mesh, D, delta, c_0)
        rho_array = f._get_rho_array(ps_centers)
        self.assertTrue(np.allclose(rho_array, rho_expected))

    def test_decay_term(self):
        dim = self.L.shape[0]
        rho_expected = 1.0 / (self.dx ** dim)

        mesh = uniform_mesh_factory(self.L, self.dx)
        r_centers = mesh.cellCenters.value.T
        ps_centers = positions.PeriodicPositions(self.L, r_centers)

        dt = 0.00001
        D = 0.0
        delta = 1.0
        c_0 = 1.0
        f = fields.FoodField(mesh, D, delta, c_0)
        f.iterate(ps_centers, dt)
        c_expected = c_0 * np.exp(-delta * rho_expected * dt)
        self.assertTrue(np.allclose(f.c, c_expected))

    def test_diff_term(self):
        dim = self.L.shape[0]

        mesh = uniform_mesh_factory(self.L, self.dx)
        r_centers = mesh.cellCenters.value.T
        r_centers_mag = np.sqrt(np.sum(np.square(r_centers), axis=-1))
        i_center = np.argmin(r_centers_mag)
        ps = positions.PeriodicPositions(self.L, np.zeros((1, dim)))

        dt = 0.01
        t_max = 0.2
        D = 1.0
        delta = 0.0
        c_0 = 1.0

        c_0_array = np.zeros(mesh.cellCenters.shape[1])
        c_0_array[i_center] = c_0

        f = fields.FoodField(mesh, D, delta, c_0_array)

        for t in np.arange(0.0, t_max, dt):
            f.iterate(ps, dt)
        variance = 2.0 * D * t_max
        mean = dim * [0.0]
        cov = np.identity(dim) * variance
        mn = multivariate_normal(mean, cov)
        c_expected = (c_0 * self.dx ** dim) * mn.pdf(r_centers)
        # print(c_expected)
        # print(f.c.value)
        # import matplotlib.pyplot as plt
        # plt.plot(r_centers_mag, f.c.value, label='Got')
        # plt.plot(r_centers_mag, c_expected, label='Expected')
        # plt.legend()
        # plt.show()
        self.assertTrue(np.allclose(f.c, c_expected, atol=1e-4))

    def test_random_seeding(self):
        dim = self.L.shape[0]
        mesh = uniform_mesh_factory(self.L, self.dx)

        dt = 0.01
        t_max = 0.2
        D = 1.0
        delta = 0.0
        c_0 = 1.0

        n = 1000
        r_0 = np.zeros([n, dim])
        ps = positions.PeriodicPositions(self.L, r_0)

        def get_f(npy_seed):
            np.random.seed(npy_seed)
            f = fields.FoodField(mesh, D, delta, c_0)
            for t in np.arange(0.0, t_max, dt):
                ps.r += self.rng.uniform(-self.dx, self.dx, size=(n, dim))
                f.iterate(ps, dt)
            return f

        f_1 = get_f(2)
        f_2 = get_f(3)

        self.assertTrue(np.allclose(f_1.c.value, f_2.c.value))


class TestFoodField2D(TestFoodField1D):
    L = np.array(2 * [4.0])
    dx = 0.05



from __future__ import print_function, division
import numpy as np
from ciabatta import crandom
from ciabatta.meta import make_repr_str
from spatious import vector


class Directions1D(object):

    def __init__(self, u_0):
        self.sign = np.sign(u_0[:, 0])
        self.sign_0 = self.sign.copy()

    @property
    def n(self):
        return self.u.shape[0]

    @property
    def dim(self):
        return self.u.shape[1]

    @property
    def u(self):
        return self.sign[:, np.newaxis].copy()

    @property
    def u_0(self):
        return self.sign_0[:, np.newaxis].copy()

    def tumble(self, tumblers, rng=None):
        if rng is None:
            rng = np.random
        self.sign[tumblers] = crandom.randbool(tumblers.sum(), rng=rng)
        return self

    def __repr__(self):
        fs = [('n', self.n)]
        return make_repr_str(self, fs)


class Directions2D(Directions1D):

    def __init__(self, u_0):
        self.th = np.arctan2(u_0[:, 1], u_0[:, 0])
        self.th_0 = self.th.copy()

    def _th_to_u(self, th):
        return np.array([np.cos(th), np.sin(th)]).T

    @property
    def u(self):
        return self._th_to_u(self.th)

    @property
    def u_0(self):
        return self._th_to_u(self.th_0)

    def tumble(self, tumblers, rng=None):
        if rng is None:
            rng = np.random
        self.th[tumblers] = rng.uniform(-np.pi, np.pi, size=tumblers.sum())
        return self

    def rotate(self, dth):
        self.th += dth
        return self

    def __repr__(self):
        fs = [('n', self.n)]
        return make_repr_str(self, fs)


def directions_nd(u_0):
    dim = u_0.shape[1]
    if dim == 1:
        return Directions1D(u_0)
    elif dim == 2:
        return Directions2D(u_0)


def directions_factory(n, dim, aligned_flag=False, rng=None):
    if aligned_flag:
        u_0 = get_aligned_vectors(n, dim)
    else:
        u_0 = get_uniform_vectors(n, dim, rng)
    return directions_nd(u_0)


def get_uniform_vectors(n, dim, rng=None):
    return vector.sphere_pick(n=n, d=dim, rng=rng)


def get_aligned_vectors(n, dim):
    u = np.zeros([n, dim])
    u[:, 0] = 1.0
    return u



from __future__ import print_function, division
import numpy as np
from ahoy import directions, c_measurers, dc_dx_measurers, noise_measurers
import test


class TestChemoNoiseMeasurer1D(test.TestBase):
    dim = 1
    L = np.array([0.7])
    noise_measurer_cls = noise_measurers.ChemoNoiseMeasurer

    def setUp(self):
        super(TestChemoNoiseMeasurer1D, self).setUp()
        self.noise_0 = 1.5
        self.v_0 = 2.2

    def run_valchemo(self, ds, chi, noise_expected):
        grad_c_measurer = c_measurers.ConstantGradCMeasurer(self.n, self.dim)
        dc_dx_measurer = dc_dx_measurers.SpatialDcDxMeasurer(ds,
                                                             grad_c_measurer)
        noise_measurer = self.noise_measurer_cls(self.noise_0, chi,
                                                 dc_dx_measurer)
        noise = noise_measurer.get_noise()
        self.assertTrue(np.allclose(noise, noise_expected))

    def get_ds_parallel(self):
        u_0 = np.zeros([self.n, self.dim])
        u_0[:, 0] = 1.0
        return directions.directions_nd(u_0)

    def get_ds_antiparallel(self):
        u_0 = np.zeros([self.n, self.dim])
        u_0[:, 0] = -1.0
        return directions.directions_nd(u_0)

    def test_parallel_nochemo(self):
        ds = self.get_ds_parallel()
        self.run_valchemo(ds, 0.0, self.noise_0)

    def test_antiparallel_nochemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 0.0, self.noise_0)

    def test_parallel_maxchemo(self):
        ds = self.get_ds_parallel()
        self.run_valchemo(ds, 1.0, 0.0)

    def test_antiparallel_maxchemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 1.0, 2.0 * self.noise_0)

    def test_parallel_halfchemo(self):
        ds = self.get_ds_parallel()
        self.run_valchemo(ds, 0.5, self.noise_0 / 2.0)

    def test_antiparallel_halfchemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 0.5, 1.5 * self.noise_0)


class TestChemoNoiseMeasurer2D(TestChemoNoiseMeasurer1D):
    dim = 2
    L = np.array([0.7, 1.1])

    def get_ds_perp(self):
        u_0 = np.zeros([self.n, self.dim])
        u_0[:, 1] = 1.0
        return directions.directions_nd(u_0)

    def test_perp_nochemo(self):
        ds = self.get_ds_perp()
        self.run_valchemo(ds, 0.0, self.noise_0)

    def test_perp_maxchemo(self):
        ds = self.get_ds_perp()
        self.run_valchemo(ds, 1.0, self.noise_0)

    def test_perp_halfchemo(self):
        ds = self.get_ds_perp()
        self.run_valchemo(ds, 0.5, self.noise_0)


class TestOneSidedChemoNoiseMeasurer1D(TestChemoNoiseMeasurer1D):
    noise_measurer_cls = noise_measurers.OneSidedChemoNoiseMeasurer

    def test_antiparallel_maxchemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 1.0, self.noise_0)

    def test_antiparallel_halfchemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 0.5, self.noise_0)


class TestOneSidedChemoRudderControllers2D(TestChemoNoiseMeasurer2D):
    noise_measurer_cls = noise_measurers.OneSidedChemoNoiseMeasurer

    def test_antiparallel_maxchemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 1.0, self.noise_0)

    def test_antiparallel_halfchemo(self):
        ds = self.get_ds_antiparallel()
        self.run_valchemo(ds, 0.5, self.noise_0)



import numpy as np
from ahoy import numerics


class RingBuffer(object):
    def __init__(self, n_p):
        self.n_p = n_p
        self.a = np.zeros([self.n_p])
        self.i_zero = 0

    def _i_dec(self):
        self.i_zero -= 1
        if self.i_zero < 0:
            self.i_zero += self.n_p

    def update(self, a_new):
        self._i_dec()
        self.a[self.i_zero] = a_new

    def integral_transform(self, K):
        tot = 0.0
        for i in range(self.n_p):
            tot += self.a[(self.i_zero + i) % self.n_p] * K[i]
        return tot


class CylinderBuffer(RingBuffer):
    def __init__(self, n_l, n_p):
        self.n_l = n_l
        self.n_p = n_p
        self.a = np.zeros([self.n_l, self.n_p])
        self.i_zero = 0

        # Optimisation
        self.inds_p = np.arange(0, self.n_p, dtype=np.uint)
        self.inds_a = self.inds_p.copy()
        self.b = np.empty([self.n_l], dtype=np.float)

    def update(self, a_new):
        self._i_dec()
        self.a[:, self.i_zero] = a_new

    def integral_transform(self, K):
        numerics.integral_transform(self.a, K, self.i_zero,
                                    self.b, self.inds_p, self.inds_a)
        return self.b



from __future__ import print_function, division
import numpy as np
from ciabatta.meta import make_repr_str
from ahoy.rudder_sets import rudder_set_factory
from ahoy.directions import directions_factory
from ahoy.positions import positions_factory
from ahoy.swimmers import swimmers_factory


class Agents(object):

    def __init__(self, directions, positions, rudder_sets, swimmers):
        self.directions = directions
        self.rudder_sets = rudder_sets
        self.positions = positions
        self.swimmers = swimmers

    def iterate(self, dt, rng, obstructor):
        self.directions = self.rudder_sets.rotate(self.directions, dt, rng)
        self.positions, dr = self.swimmers.displace(self.positions, dt)
        obstructor.obstruct(self.positions, dr, self.directions)

    @property
    def chi(self):
        return self.rudder_sets.chi

    @property
    def n(self):
        return self.directions.n

    def __repr__(self):
        fs = [('directions', self.directions), ('positions', self.positions),
              ('rudder_sets', self.rudder_sets), ('swimmers', self.swimmers)]
        return make_repr_str(self, fs)


def agents_factory(rng, dim, aligned_flag,
                   n=None, rho_0=None,
                   chi=None, onesided_flag=None,
                   tumble_flag=None, p_0=None, tumble_chemo_flag=None,
                   rotation_flag=None, Dr_0=None, rotation_chemo_flag=None,
                   temporal_chemo_flag=None, dt_mem=None, t_mem=None, time=None,
                   spatial_flag=None, v_0=None,
                   periodic_flag=None, L=None, origin_flags=None, obstructor=None,
                   c_field_flag=None, c_field=None):
    if rho_0 is not None:
        try:
            volume_free = obstructor.volume_free
        except AttributeError:
            volume_free = np.product(L)
        n = int(round(rho_0 * volume_free))
    ds = directions_factory(n, dim, aligned_flag=aligned_flag,
                            rng=rng)
    ps = positions_factory(spatial_flag, periodic_flag, n, dim, L,
                           origin_flags, rng, obstructor)
    rudder_sets = rudder_set_factory(temporal_chemo_flag,
                                     ds,
                                     ps, v_0, dt_mem, t_mem, time,
                                     c_field_flag, c_field,
                                     onesided_flag, chi,
                                     tumble_flag, p_0, tumble_chemo_flag,
                                     rotation_flag, Dr_0, dim, rotation_chemo_flag)
    swims = swimmers_factory(spatial_flag, v_0, ds)
    return Agents(ds, ps, rudder_sets, swims)



from __future__ import print_function, division
from ciabatta.meta import make_repr_str


class Time(object):
    def __init__(self):
        self.t = 0.0

    def iterate(self, dt):
        self.t += dt

    def __repr__(self):
        fs = [('t', self.t)]
        return make_repr_str(self, fs)



import numpy as np
from ahoy import mesh
import test


class TestUniformMesh1D(test.TestBase):
    dim = 1
    L = np.array([1.8])
    dx = 0.1

    def test_uniform_mesh(self):
        msh = mesh.uniform_mesh_factory(self.L, self.dx)
        for i_dim in range(self.dim):
            self.assertTrue(msh.cellCenters[i_dim, :].value.max() <
                            self.L[i_dim] / 2.0)
            self.assertTrue(msh.cellCenters[i_dim, :].value.min() >
                            -self.L[i_dim] / 2.0)


class TestUniformMesh2D(TestUniformMesh1D):
    dim = 2
    L = np.array([1.7, 3.0])


class TestPorousMesh(test.TestBase):

    def test_porous_random_seeding(self):
        '''Test the reproducibility of the meshing algorithm.'''

        L = np.array([2.0, 2.0])
        R = 0.1
        dx = 0.1
        rs = np.array([
            [-0.5, 0.5],
            [0.0, 0.0],
            [0.1, 0.4],
            [0.88, 0.88],
            [0.88, 0.0],
        ])

        np.random.seed(2)
        mesh_1 = mesh.porous_mesh_factory(rs, R, dx, L)

        np.random.seed(3)
        mesh_2 = mesh.porous_mesh_factory(rs, R, dx, L)

        self.assertTrue(np.allclose(mesh_1.cellCenters.value,
                                    mesh_2.cellCenters.value))



from __future__ import print_function, division
import numpy as np
from spatious import vector
from agaro.output_utils import (get_filenames, filename_to_model,
                                get_recent_model)
from agaro.measure_utils import measures
from ahoy import rudders


def seg_intersect(p1, p2, yi):
    x1, y1 = p1
    x2, y2 = p2
    m = (y2 - y1) / (x2 - x1)
    c = y1 - m * x1
    xi = (yi - c) / m
    if x1 < xi < x2:
        return xi
    else:
        raise ValueError


def curve_intersect(xs, ys, yi):
    i_big = np.where(ys > yi)[0][0]
    p1 = [xs[i_big - 1], ys[i_big - 1]]
    p2 = [xs[i_big], ys[i_big]]
    return seg_intersect(p1, p2, yi)


def get_vd_coeff(x, t):
    return x / t


def get_diff_coeff(x, t):
    return x ** 2 / (2.0 * t)


def get_ud_vector(m):
    dr = m.ships.agents.positions.dr
    return (np.mean(get_vd_coeff(dr, m.ships.time.t), axis=0) /
            m.ships.agents.swimmers.v_0)


def get_ud_scalar(m):
    dr = m.ships.agents.positions.dr_mag
    return (np.mean(get_vd_coeff(dr, m.ships.time.t), axis=0) /
            m.ships.agents.swimmers.v_0)


def get_D_vector(m):
    dr = m.ships.agents.positions.dr
    return np.mean(get_diff_coeff(dr, m.ships.time.t), axis=0)


def get_D_scalar(m):
    dr = m.ships.agents.positions.dr_mag
    return np.mean(get_diff_coeff(dr, m.ships.time.t), axis=0)


def get_r_vector(m):
    dr = m.ships.agents.positions.dr
    return np.mean(dr, axis=0)


def get_r_scalar(m):
    dr = m.ships.agents.positions.dr_mag
    return np.mean(dr, axis=0)


def get_u_net_vector(m):
    return np.mean(m.ships.agents.directions.u, axis=0)


def get_u_net_scalar(m):
    return vector.vector_mag(get_u_net_vector(m))


def get_chi(m):
    return m.ships.agents.chi


def get_pf(m):
    return m.ships.obstructor.fraction_occupied


def get_Dr_0(m):
    for ruds in m.ships.agents.rudder_sets:
        if isinstance(ruds, rudders.RotationRudders):
            return ruds.noise_measurer.noise_0
    return 0.0


def get_p_0(m):
    for ruds in m.ships.agents.rudder_sets:
        if isinstance(ruds, rudders.TumbleRudders):
            return ruds.noise_measurer.noise_0
    return 0.0


def _t_measures(dirname, measure_func):
    ts, measures = [], []
    for fname in get_filenames(dirname):
        m = filename_to_model(fname)
        ts.append(m.ships.time.t)
        measures.append(measure_func(m))
    return np.array(ts), np.array(measures)


def t_uds_vector(dirname):
    """Calculate the particle drift speed over time along each axis
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    uds: numpy.ndarray[dtype=float]
         Drift speeds, normalised by the swimmer speed.
    """
    return _t_measures(dirname, get_ud_vector)


def t_uds_scalar(dirname):
    """Calculate the overall particle drift speed over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    uds: numpy.ndarray[dtype=float]
         Particle drift speeds.
    """
    return _t_measures(dirname, get_ud_scalar)


def t_Ds_scalar(dirname):
    """Calculate the overall particle diffusion constant over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    Ds: numpy.ndarray[dtype=float]
         Particle diffusion constants.
    """
    return _t_measures(dirname, get_D_scalar)


def t_Ds_vector(dirname):
    """Calculate the particle diffusion constant over time along each axis
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    Ds: numpy.ndarray[dtype=float]
         Particle diffusion constants.
    """
    return _t_measures(dirname, get_D_vector)


def t_rs_scalar(dirname):
    """Calculate the overall particle displacement over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    rs: numpy.ndarray[dtype=float]
         Particle diffusion constants.
    """
    return _t_measures(dirname, get_r_scalar)


def t_rs_vector(dirname):
    """Calculate the particle displacement over time along each axis
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    rs: numpy.ndarray[dtype=float]
         Particle diffusion constants.
    """
    return _t_measures(dirname, get_r_vector)


def t_u_nets_scalar(dirname):
    """Calculate the particles' overall centre-of-mass speed over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    v_nets: numpy.ndarray[dtype=float]
         Centre-of-mass particle speeds.
    """
    return _t_measures(dirname, get_u_net_scalar)


def t_u_nets_vector(dirname):
    """Calculate the particle's centre-of-mass velocity over time
    for a model output directory.

    Parameters
    ----------
    dirname: str
        A model output directory path

    Returns
    -------
    ts: numpy.ndarray[dtype=float]
        Times.
    v_nets: numpy.ndarray[dtype=float]
         Centre-of-mass particle velocities.
    """
    return _t_measures(dirname, get_u_net_vector)


def chi_uds_x(dirnames, t_steady=None):
    chis = measures(dirnames, get_chi, t_steady)
    uds = measures(dirnames, get_ud_vector, t_steady)
    return chis, uds[:, 0]


def pf_Ds_scalar(dirnames, t_steady=None):
    pfs = measures(dirnames, get_pf, t_steady)
    Ds = measures(dirnames, get_D_scalar, t_steady)
    return pfs, Ds


def Dr_0_Ds_scalar(dirnames, t_steady=None):
    Dr_0s = measures(dirnames, get_Dr_0, t_steady)
    Ds = measures(dirnames, get_D_scalar, t_steady)
    return Dr_0s, Ds


def p_0_Ds_scalar(dirnames, t_steady=None):
    p_0s = measures(dirnames, get_p_0, t_steady)
    Ds = measures(dirnames, get_D_scalar, t_steady)
    return p_0s, Ds


def pf_uds_x(dirnames, t_steady=None):
    pfs = measures(dirnames, get_pf, t_steady)
    uds = measures(dirnames, get_ud_vector, t_steady)
    return pfs, uds[:, 0]


def get_equiv_chi(ud_0, dirnames):
    chis, uds = chi_uds_x(dirnames)
    i_sort = np.argsort(chis)
    chis, uds = chis[i_sort], uds[i_sort]
    return curve_intersect(chis, uds, ud_0)


def get_equiv_chi_key(m):
    noise_var_key = 'p_0' if m.ships.agents.does_tumbling else 'Dr_0'
    chemo_rudders = m.ships.agents.rudder_sets.chemo_rudders
    key = (noise_var_key, chemo_rudders.is_onesided,
           chemo_rudders.noise_measurer.is_temporal)
    return key


def get_equiv_chi_item(ud_0, dirnames):
    key = get_equiv_chi_key(get_recent_model(dirnames[0]))
    chi_equiv = get_equiv_chi(ud_0, dirnames)
    return key, chi_equiv


def get_equiv_chi_dict(ud_0, dirname_sets):
    params_to_chi = {}
    for dirnames in dirname_sets:
        key, chi_equiv = get_equiv_chi_item(ud_0, dirnames)
        params_to_chi[key] = chi_equiv
    return params_to_chi


def circle_segment_angle(d, R):
    return 2.0 * np.arccos(d / R)


def circle_segment_area(d, R):
    if d > R:
        return 0.0
    elif d < -R:
        return np.pi * R ** 2.0
    else:
        theta = circle_segment_angle(d, R)
        return ((R ** 2) / 2.0) * (theta - np.sin(theta))


def circle_cross_section_area(d_1, d_2, R):
    return np.abs(circle_segment_area(d_1, R) - circle_segment_area(d_2, R))


def linear_areas(rs, R, Lx, Ly, nx):
    x = np.linspace(-Lx / 2.0, Lx / 2.0, nx)
    linear_areas = np.full_like(x, Lx * Ly / nx)
    for i_r in range(rs.shape[0]):
        for i in range(nx - 1):
            d_1 = rs[i_r] - x[i]
            d_2 = rs[i_r] - x[i + 1]
            cross_section_area = circle_cross_section_area(d_1, d_2, R)
            linear_areas[i] -= cross_section_area
    return linear_areas


def linear_density(xs, xcs, R, Lx, Ly, dx):
    nx = int(round(Lx / dx))
    ns, x_bins = np.histogram(xs, bins=nx, range=(-Lx / 2.0, Lx / 2.0))
    areas = linear_areas(xcs, R, Lx, Ly, nx)
    densities = ns.astype(np.float) / areas
    return densities, x_bins


def get_linear_density(m, dx):
    xs = m.ships.agents.positions.r[:, 0]
    try:
        xcs = m.ships.obstructor.rs[:, 0]
        R = m.ships.obstructor.R
    except AttributeError:
        xcs = np.array([])
        R = 0.0
    Lx, Ly = m.ships.agents.positions.L
    return linear_density(xs, xcs, R, Lx, Ly, dx)


def angle_density(ths, dth):
    nth = int(round(2.0 * np.pi / dth))
    ns, th_bins = np.histogram(ths, bins=nth, range=(-np.pi, np.pi))
    dth = th_bins[1] - th_bins[0]
    densities = ns.astype(np.float) / (ns * dth).sum()
    return densities, th_bins


def get_angle_density(m, dth):
    ths = m.ships.agents.directions.th
    return angle_density(ths, dth)



from __future__ import print_function, division
from ciabatta.meta import make_repr_str


class Swimmers(object):

    def __init__(self, v_0, directions):
        self.v_0 = v_0
        self.directions = directions

    def displace(self, positions, dt):
        dr = self.v_0 * self.directions.u * dt
        positions.r += dr
        return positions, dr

    def __repr__(self):
        fs = [('v_0', self.v_0)]
        return make_repr_str(self, fs)


class NoneSwimmers(object):

    def displace(self, positions, dt):
        dr = None
        return positions, dr


def swimmers_factory(spatial_flag, v_0, ds):
    if spatial_flag:
        return Swimmers(v_0, ds)
    else:
        return NoneSwimmers()



from __future__ import print_function, division
from abc import ABCMeta, abstractmethod
import numpy as np
from ciabatta.meta import make_repr_str
from ahoy import noise_measurers
from ahoy.noise_measurers import noise_measurer_factory


class Rudders(object):

    def __init__(self, noise_measurer):
        self.noise_measurer = noise_measurer

    @property
    def is_chemotactic(self):
        return isinstance(self.noise_measurer,
                          noise_measurers.ChemoNoiseMeasurer)

    @property
    def is_onesided(self):
        return isinstance(self.noise_measurer,
                          noise_measurers.OneSidedChemoNoiseMeasurer)

    @property
    def chi(self):
        if self.is_chemotactic:
            return self.noise_measurer.chi
        else:
            return None

    def rotate(self, directions, dt, rng=None):
        noise = self.noise_measurer.get_noise()
        return self._rotate(directions, noise, dt, rng)

    def _rotate(self, directions, noise, dt, rng):
        return directions

    def __repr__(self):
        fs = [('noise_measurer', self.noise_measurer)]
        return make_repr_str(self, fs)


class RotationRudders(Rudders):
    __metaclass__ = ABCMeta

    @abstractmethod
    def _get_dth(self, directions, noise, dt):
        return

    def _rotate(self, directions, noise, dt, rng):
        dth = self._get_dth(directions, noise, dt, rng)
        return directions.rotate(dth)


class RotationRudders2D(RotationRudders):

    def _get_dth(self, directions, noise, dt, rng):
        if rng is None:
            rng = np.random
        return rng.normal(scale=np.sqrt(2.0 * noise * dt), size=directions.n)


class TumbleRudders(Rudders):

    def _get_tumblers(self, directions, noise, dt, rng):
        if rng is None:
            rng = np.random
        return rng.uniform(size=directions.n) < noise * dt

    def _rotate(self, directions, noise, dt, rng):
        tumblers = self._get_tumblers(directions, noise, dt, rng)
        return directions.tumble(tumblers, rng)


def rotation_rudders_nd(dim, *args, **kwargs):
    if dim == 2:
        return RotationRudders2D(*args, **kwargs)
    else:
        raise NotImplementedError('No rotation rudders implemented in this '
                                  ' dimension')


def rudders_factory(tumble_flag, dim,
                    chemo_flag,
                    noise_0,
                    onesided_flag, chi,
                    temporal_chemo_flag,
                    ds,
                    ps, v_0, dt_mem, t_mem, t_rot_0, time,
                    c_field_flag, c_field):
    noise_measurer = noise_measurer_factory(chemo_flag,
                                            noise_0,
                                            onesided_flag, chi,
                                            temporal_chemo_flag,
                                            ds,
                                            ps, v_0, dt_mem, t_mem, t_rot_0, time,
                                            c_field_flag, c_field)
    if tumble_flag:
        rudders = TumbleRudders(noise_measurer)
    else:
        rudders = rotation_rudders_nd(dim, noise_measurer)
    return rudders



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools
from setuptools import setup, Extension
from Cython.Build import cythonize
import numpy

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    'Cython',
    'numpy',
    'scipy',
    'matplotlib',
    'ez_setup',
    'fipy',
    'ciabatta',
    'spatious',
    'fealty',
    'agaro',
    'metropack',
]

test_requirements = [
    'Cython',
    'numpy',
    'scipy',
    'matplotlib',
    'ez_setup',
    'fipy',
    'ciabatta',
    'spatious',
    'fealty',
    'agaro',
    'metropack',
]

extensions = cythonize([
    Extension("ahoy.numerics", ["ahoy/numerics.pyx"],
              include_dirs=[numpy.get_include()]),
])

console_scripts = [
    'plot_chi_uds_x = ahoy.utils.scripts:plot_chi_uds_x',
    'plot_Dr_0_Ds_scalar = ahoy.utils.scripts:plot_Dr_0_Ds_scalar',
    'plot_linear_density = ahoy.utils.scripts:plot_linear_density',
    'plot_p_0_Ds_scalar = ahoy.utils.scripts:plot_p_0_Ds_scalar',
    'plot_pf_Ds_scalar = ahoy.utils.scripts:plot_pf_Ds_scalar',
    'plot_pf_uds_x = ahoy.utils.scripts:plot_pf_uds_x',
    'plot_t_Ds_scalar = ahoy.utils.scripts:plot_t_Ds_scalar',
    'plot_t_Ds_vector = ahoy.utils.scripts:plot_t_Ds_vector',
    'plot_t_rs_scalar = ahoy.utils.scripts:plot_t_rs_scalar',
    'plot_t_rs_vector = ahoy.utils.scripts:plot_t_rs_vector',
    'plot_t_u_nets_scalar = ahoy.utils.scripts:plot_t_u_nets_scalar',
    'plot_t_u_nets_vector = ahoy.utils.scripts:plot_t_u_nets_vector',
    'plot_t_uds_scalar = ahoy.utils.scripts:plot_t_uds_scalar',
    'plot_t_uds_vector = ahoy.utils.scripts:plot_t_uds_vector',
    'plot_vis = ahoy.utils.scripts:plot_vis',
]

setup(
    name='ahoy',
    version='0.4.2',
    description="Agent-based simulations of active particles",
    long_description=readme + '\n\n' + history,
    author="Elliot Marsden",
    author_email='elliot.marsden@gmail.com',
    url='https://github.com/eddiejessup/ahoy',
    packages=setuptools.find_packages(exclude=['docs', 'tests']),
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='ahoy',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    ext_modules=extensions,
    entry_points={
        'console_scripts': console_scripts,
    }
)



from __future__ import print_function, division
import numpy as np
from ciabatta.meta import make_repr_str
from ahoy import measurers
from ahoy.dc_dx_measurers import TemporalDcDxMeasurer, dc_dx_factory


class NoiseMeasurer(measurers.Measurer):

    def __init__(self, noise_0, *args, **kwargs):
        self.noise_0 = noise_0

    def get_noise(self):
        return self.noise_0

    def __repr__(self):
        fs = [('noise_0', self.noise_0)]
        return make_repr_str(self, fs)


class ChemoNoiseMeasurer(NoiseMeasurer):

    def __init__(self, noise_0, chi, dc_dx_measurer):
        NoiseMeasurer.__init__(self, noise_0)
        self.chi = chi
        self.dc_dx_measurer = dc_dx_measurer

    @property
    def is_temporal(self):
        return isinstance(self.dc_dx_measurer, TemporalDcDxMeasurer)

    def get_noise(self):
        dc_dxs = self.dc_dx_measurer.get_dc_dxs()
        return self.noise_0 * (1.0 - self.chi * dc_dxs)

    def __repr__(self):
        fs = [('noise_0', self.noise_0), ('chi', self.chi),
              ('dc_dx_measurer', self.dc_dx_measurer)]
        return make_repr_str(self, fs)


class OneSidedChemoNoiseMeasurer(ChemoNoiseMeasurer):

    def get_noise(self):
        noise_two_sided = super(OneSidedChemoNoiseMeasurer, self).get_noise()
        return np.minimum(self.noise_0, noise_two_sided)


def chemo_noise_measurer_factory(onesided_flag, *args, **kwargs):
    if onesided_flag:
        return OneSidedChemoNoiseMeasurer(*args, **kwargs)
    else:
        return ChemoNoiseMeasurer(*args, **kwargs)


def noise_measurer_factory(chemo_flag,
                           noise_0,
                           onesided_flag, chi,
                           temporal_chemo_flag,
                           ds,
                           ps, v_0, dt_mem, t_mem, t_rot_0, time,
                           c_field_flag, c_field):
    if chemo_flag:
        dc_dx_measurer = dc_dx_factory(temporal_chemo_flag,
                                       ds,
                                       ps, v_0, dt_mem, t_mem, t_rot_0, time,
                                       c_field_flag, c_field)
        return chemo_noise_measurer_factory(onesided_flag, noise_0, chi,
                                            dc_dx_measurer)
    else:
        return NoiseMeasurer(noise_0)



from __future__ import print_function, division
from abc import ABCMeta, abstractmethod
import numpy as np
from ciabatta.meta import make_repr_str
from spatious import vector, distance, geom
from metropack import pack
from ahoy import mesh, turners


class NoneObstructor(object):

    def get_obstructeds(self, rs):
        return np.zeros([rs.shape[0]], dtype=np.bool)

    def obstruct(self, ps, drs, ds):
        return

    def get_mesh(self, L, dx):
        return mesh.uniform_mesh_factory(L, dx)

    def __repr__(self):
        fs = []
        return make_repr_str(self, fs)


class BaseObstructor(NoneObstructor):
    __metaclass__ = ABCMeta

    def __init__(self, turner):
        self.turner = turner

    def _push(self, obs, rs, drs):
        rs[obs] -= drs[obs]

    def get_obstructeds(self, rs):
        seps = self.get_seps(rs)
        return self._is_obstructed(seps)


class SphereObstructor(BaseObstructor):
    __metaclass__ = ABCMeta

    def __init__(self, turner, R, *args, **kwargs):
        super(SphereObstructor, self).__init__(turner)
        self.R = R

    def _is_obstructed(self, seps):
        return vector.vector_mag_sq(seps) < self.R ** 2.0

    @property
    def volume_sphere(self):
        return geom.sphere_volume(self.R, self.dim)

    @abstractmethod
    def _get_th_normals(self, seps):
        return

    @abstractmethod
    def get_seps(self, rs):
        return rs

    def obstruct(self, ps, drs, ds, rng=None):
        seps = self.get_seps(ps.r_w)
        obs = self._is_obstructed(seps)
        self._push(obs, ps.r, drs)
        th_normals = self._get_th_normals(seps[obs])
        self.turner.turn(obs, ds, th_normals, rng)

    def __repr__(self):
        fs = [('turner', self.turner), ('R', self.R)]
        return make_repr_str(self, fs)


class SphereObstructor2D(SphereObstructor):

    def _get_th_normals(self, seps):
        return np.arctan2(seps[:, 1], seps[:, 0])


class SingleSphereObstructor2D(SphereObstructor2D):

    def get_seps(self, rs):
        return rs

    def get_mesh(self, L, dx):
        raise NotImplementedError


class PorousObstructor(SphereObstructor2D):

    def __init__(self, turner, R, L, pf, rng, periodic_flag):
        super(PorousObstructor, self).__init__(turner, R)
        self.periodic_flag = periodic_flag
        self.L = L
        self.rs, self.R = pack.pack(self.R, self.L, pf=pf, rng=rng,
                                    periodic=periodic_flag)

    def obstruct(self, *args, **kwargs):
        if self.rs.shape[0]:
            super(PorousObstructor, self).obstruct(*args, **kwargs)

    def get_obstructeds(self, *args, **kwargs):
        if self.rs.shape[0]:
            return super(PorousObstructor, self).get_obstructeds(*args,
                                                                 **kwargs)
        else:
            return NoneObstructor.get_obstructeds(self, *args, **kwargs)

    @property
    def volume(self):
        return np.product(self.L)

    @property
    def dim(self):
        return self.L.shape[0]

    @property
    def n(self):
        return self.rs.shape[0]

    @property
    def volume_occupied(self):
        return self.n * self.volume_sphere

    @property
    def fraction_occupied(self):
        return self.volume_occupied / self.volume

    @property
    def volume_free(self):
        return self.volume - self.volume_occupied

    @property
    def fraction_free(self):
        return self.volume_free / self.volume

    def get_seps(self, rs):
        return distance.csep_periodic_close(rs, self.rs, self.L)[0]

    def get_mesh(self, L, dx):
        return mesh.porous_mesh_factory(self.rs, self.R, dx, L)

    def __repr__(self):
        fs = [('turner', self.turner), ('R', self.R), ('L', self.L),
              ('fraction_occupied', self.fraction_occupied)]
        return make_repr_str(self, fs)

turner_str_map = {
    'stall': turners.Turner(),
    'bounce_back': turners.BounceBackTurner(),
    'reflect': turners.ReflectTurner(),
    'align': turners.AlignTurner(),
}


def obstructor_factory(pore_flag, turner_str, R, L, pf, rng, periodic_flag):
    if pore_flag:
        return PorousObstructor(turner_str_map[turner_str], R, L, pf, rng,
                                periodic_flag)
    else:
        return NoneObstructor()



from __future__ import print_function, division
from abc import ABCMeta, abstractmethod
import numpy as np
from ciabatta.meta import make_repr_str
from ahoy import measurers


class CMeasurer(measurers.Measurer):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_cs(self):
        return


class FieldCMeasurer(CMeasurer):

    def __init__(self, c_field, positions):
        self.c_field = c_field
        self.positions = positions

    def get_cs(self):
        return self.c_field.get_val_i(self.positions)

    def __repr__(self):
        fs = [('c_field', self.c_field)]
        return make_repr_str(self, fs)


class LinearCMeasurer(CMeasurer):

    def __init__(self, positions):
        self.positions = positions

    def get_cs(self):
        return self.positions.dr[:, 0]

    def __repr__(self):
        fs = []
        return make_repr_str(self, fs)


class GradCMeasurer(measurers.Measurer):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_grad_cs(self):
        return


class FieldGradCMeasurer(GradCMeasurer):

    def __init__(self, c_field, positions):
        self.c_field = c_field
        self.positions = positions

    def get_grad_cs(self):
        return self.c_field.get_grad_i(self.positions)

    def __repr__(self):
        fs = [('c_field', self.c_field)]
        return make_repr_str(self, fs)


class ConstantGradCMeasurer(GradCMeasurer):

    def __init__(self, n, dim):
        self.grad_c = np.zeros([n, dim])
        self.grad_c[:, 0] = 1.0

    @property
    def n(self):
        return self.grad_c.shape[0]

    @property
    def dim(self):
        return self.grad_c.shape[1]

    def get_grad_cs(self):
        return self.grad_c

    def __repr__(self):
        fs = []
        return make_repr_str(self, fs)



from __future__ import print_function, division
from ciabatta.meta import make_repr_str
from ahoy.stime import Time
from ahoy.obstructors import obstructor_factory
from ahoy.agents import agents_factory
from ahoy.fields import food_field_factory


class Ships(object):

    def __init__(self, time, ags, obstructor, c_field):
        self.time = time
        self.agents = ags
        self.obstructor = obstructor
        self.c_field = c_field

    @property
    def dim(self):
        return self.agents.directions.dim

    @property
    def t(self):
        return self.time.t

    def iterate(self, dt, rng):
        self.agents.iterate(dt, rng, self.obstructor)
        self.c_field.iterate(self.agents.positions, dt)
        self.time.iterate(dt)

    def __repr__(self):
        fs = [('time', self.time), ('agents', self.agents),
              ('obstructor', self.obstructor), ('c_field', self.c_field)]
        return make_repr_str(self, fs)


def ships_factory(rng, dim,
                  aligned_flag=None,
                  n=None, rho_0=None,
                  spatial_flag=None, v_0=None,
                  periodic_flag=None, L=None, origin_flags=None,
                  chi=None, onesided_flag=None,
                  tumble_flag=None, p_0=None, tumble_chemo_flag=None,
                  rotation_flag=None, Dr_0=None, rotation_chemo_flag=None,
                  temporal_chemo_flag=None, dt_mem=None, t_mem=None,
                  pore_flag=None, pore_turner=None, pore_R=None, pore_pf=None,
                  c_field_flag=None, c_dx=None, c_D=None, c_delta=None,
                  c_0=None):
    time = Time()
    periodic_flag = not c_field_flag
    obstructor = obstructor_factory(pore_flag, pore_turner,
                                    pore_R, L, pore_pf, rng,
                                    periodic_flag)
    c_field = food_field_factory(c_field_flag, L, c_dx, c_D, c_delta,
                                 c_0, obstructor)
    ags = agents_factory(rng, dim, aligned_flag,
                         n, rho_0,
                         chi, onesided_flag,
                         tumble_flag, p_0, tumble_chemo_flag,
                         rotation_flag, Dr_0, rotation_chemo_flag,
                         temporal_chemo_flag, dt_mem, t_mem, time,
                         spatial_flag, v_0,
                         periodic_flag, L, origin_flags, obstructor,
                         c_field_flag, c_field)
    return Ships(time, ags, obstructor, c_field)



from __future__ import print_function, division
import numpy as np
from ahoy import directions
import test


class TestDirections1D(test.TestBase):
    dim = 1

    def test_properties(self):
        n = 10
        u = np.zeros([n, self.dim])
        u[:, 0] = 1.0
        ds = directions.directions_nd(u)
        self.assertTrue(ds.n == n)
        self.assertTrue(ds.dim == self.dim)

    def test_u_east(self):
        u = np.zeros([self.n, self.dim])
        u[:, 0] = 1.0
        ds = directions.directions_nd(u)
        self.assertTrue(np.allclose(ds.u, u))

    def test_u_west(self):
        u = np.zeros([self.n, self.dim])
        u[:, 0] = -1.0
        ds = directions.directions_nd(u)
        self.assertTrue(np.allclose(ds.u, u))

    def test_tumble_identity(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        tumblers = np.zeros([ds.n], dtype=np.bool)
        ds_rot = ds.tumble(tumblers)
        self.assertTrue(np.allclose(ds.u_0, ds_rot.u))

    def test_tumble_magnitude_conservation(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        mags_0 = np.sum(np.square(ds.u))
        tumblers = self.rng.choice([True, False], size=ds.n)
        ds_rot = ds.tumble(tumblers)
        mags_rot = np.sum(np.square(ds_rot.u))
        self.assertTrue(np.allclose(mags_0, mags_rot))

    def test_tumble_coverage(self):
        n = 2000
        ds = directions.directions_factory(n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        tumblers = np.ones([ds.n], dtype=np.bool)
        ds_rot = ds.tumble(tumblers)
        frac_close = np.isclose(ds.u_0, ds_rot.u).sum() / float(ds.n)
        self.assertAlmostEqual(frac_close, 0.5, 1)


class TestDirections2D(TestDirections1D):
    dim = 2

    def test_u_south(self):
        u = np.zeros([self.n, self.dim])
        u[:, 1] = -1.0
        ds = directions.directions_nd(u)
        self.assertTrue(np.allclose(ds.u, u))

    def test_tumble_coverage(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        tumblers = np.ones([ds.n], dtype=np.bool)
        ds_rot = ds.tumble(tumblers, rng=self.rng)
        self.assertFalse(np.any(np.isclose(ds.u_0, ds_rot.u)))

    def test_rotate_identity(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        dth = np.zeros([ds.n])
        ds_rot = ds.rotate(dth)
        u_rot = ds_rot.u
        self.assertTrue(np.allclose(ds.u_0, u_rot))

    def test_rotate_right_angle(self):
        u_0 = np.zeros([self.n, self.dim])
        u_0[:, 0] = 1.0
        dth = np.full([self.n], np.pi / 2.0)
        ds = directions.directions_nd(u_0)
        ds_rot = ds.rotate(dth)
        u_rot = ds_rot.u
        u_rot_expected = np.zeros([self.n, self.dim])
        u_rot_expected[:, 1] = 1.0
        self.assertTrue(np.allclose(u_rot, u_rot_expected))

    def test_rotate_idempotence(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        dth = self.rng.uniform(-np.pi, np.pi, size=self.n)
        ds_rot = ds.rotate(dth)
        ds_rot = ds_rot.rotate(-dth)
        self.assertTrue(np.allclose(ds.u_0, ds_rot.u))

    def test_rotate_periodicity(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        dth = np.full([ds.n], np.pi / 2.0)
        ds_rot = ds
        for i in range(4):
            ds_rot = ds_rot.rotate(dth)
        self.assertTrue(np.allclose(ds.u_0, ds_rot.u))

    def test_rotate_magnitude_conservation(self):
        ds = directions.directions_factory(self.n, self.dim,
                                           aligned_flag=False, rng=self.rng)
        mags_0 = np.sum(np.square(ds.u))
        dth = self.rng.uniform(-np.pi, np.pi, size=self.n)
        ds_rot = ds.rotate(dth)
        mags_rot = np.sum(np.square(ds_rot.u))
        self.assertTrue(np.allclose(mags_0, mags_rot))


class TestDirectionsFactories(test.TestBase):

    def test_uniform_directions_isotropy(self):
        n = 1e5
        for dim in [1, 2]:
            u_0 = directions.get_uniform_vectors(n, dim, self.rng)
            u_net = np.mean(u_0, axis=0)
            u_net_mag = np.sqrt(np.sum(np.square(u_net)))
            self.assertTrue(u_net_mag < 1e-2)

    def test_uniform_directions_magnitude(self):
        n = 1e5
        for dim in [1, 2]:
            u_0 = directions.get_uniform_vectors(n, dim, self.rng)
            u_mags = np.sum(np.square(u_0), axis=-1)
            self.assertTrue(np.allclose(u_mags, 1.0))



# -*- coding: utf-8 -*-

__author__ = 'Elliot Marsden'
__email__ = 'elliot.marsden@gmail.com'
__version__ = '0.4.2'

from ahoy import directions
from ahoy import positions
from ahoy import measurers
from ahoy import rudders
from ahoy import swimmers
from ahoy import agents
from ahoy import ships
from ahoy import stime



import numpy as np
from scipy.optimize import curve_fit
from ahoy import directions, rudders, noise_measurers
import test


class TestRotationRudders2D(test.TestBase):
    rudders_cls = rudders.RotationRudders2D
    dim = 2

    def setUp(self):
        super(TestRotationRudders2D, self).setUp()
        self.dt = 0.005
        self.noise_0 = 2.0
        self.t_rot_expect = 1.0 / self.noise_0
        self.noise_measurer = noise_measurers.NoiseMeasurer(self.noise_0)

    def test_rudder_autocorrelation(self):
        n = 10000
        t_max = 0.5

        ds = directions.directions_factory(n, self.dim, aligned_flag=True)
        rudders = self.rudders_cls(self.noise_measurer)

        u_0 = ds.u
        ts = np.arange(0.0, t_max, self.dt)
        mean_dots = []
        for t in ts:
            ds = rudders.rotate(ds, self.dt, self.rng)
            mean_dot = np.mean(np.sum(u_0 * ds.u, axis=-1))
            mean_dots.append(mean_dot)

        def exp(t, t_rot):
            return np.exp(-t / t_rot)

        popt, pcov = curve_fit(exp, ts, mean_dots)
        t_rot_actual = popt[0]
        self.assertAlmostEqual(t_rot_actual, self.t_rot_expect, 1)

    def test_random_seeding(self):
        n = 1000
        num_iterations = 100
        rng_seed = 1

        def get_ds(npy_seed):
            rng = np.random.RandomState(rng_seed)
            np.random.seed(npy_seed)
            ds = directions.directions_factory(n, self.dim, aligned_flag=True)
            ruds = self.rudders_cls(self.noise_measurer)
            for _ in range(num_iterations):
                ds = ruds.rotate(ds, self.dt, rng)
            return ds

        ds_1 = get_ds(2)
        ds_2 = get_ds(3)

        self.assertTrue(np.allclose(ds_1.u, ds_2.u))


class TestTumbleRudders1D(TestRotationRudders2D):
    rudders_cls = rudders.TumbleRudders
    dim = 1

    def test_tumble_rate(self):
        n = 2000
        dt = 1.0
        p = self.rng.uniform(0.0, 1.0, size=n)
        noise_measurer = noise_measurers.NoiseMeasurer(p)

        n_expected = (p * dt).sum()
        # In 1D is a half chance to tumble back to the same direction.
        if self.dim == 1:
            n_expected /= 2.0
        ds = directions.directions_factory(n, self.dim, aligned_flag=False,
                                           rng=self.rng)
        u_0 = ds.u
        ruds = rudders.TumbleRudders(noise_measurer)
        ds_rot = ruds.rotate(ds, dt, self.rng)
        u_change = np.not_equal(u_0, ds_rot.u)
        n_actual = np.any(u_change, axis=-1).sum()
        err = np.abs(n_expected - n_actual) / n_expected
        self.assertTrue(err < 0.05)


class TestTumbleRudders2D(TestTumbleRudders1D):
    dim = 2



import numpy as np
cimport numpy as np
cimport cython


@cython.cdivision(True)
@cython.boundscheck(False)
def integral_transform(np.ndarray[np.float_t, ndim=2] a,
                       np.ndarray[np.float_t, ndim=1] K,
                       unsigned int i_zero,
                       np.ndarray[np.float_t, ndim=1] b,
                       np.ndarray[np.uint_t, ndim=1] inds_p,
                       np.ndarray[np.uint_t, ndim=1] inds_a):
    cdef:
        unsigned int n_l = a.shape[0]
        unsigned int n_p = a.shape[1]
        unsigned int i_l, i_p
        double tot
    np.mod(i_zero + inds_p, n_p, inds_a)

    for i_l in range(n_l):
        tot = 0.0
        for i_p in range(n_p):
            tot += a[i_l, inds_a[i_p]] * K[i_p]
        b[i_l] = tot
